#!/usr/bin/env python

import pydetect

import numpy as np
import pandas as pd
import pandas.util.testing as tm


class TestGeneralizedESD(tm.TestCase):

    def test_sample(self):
        """
        the test is derived from
        http://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm
        """

        d = pydetect.GESDDetector(alpha=0.05)
        self.assertEqual(d._get_lambda(54, 0), 3.1587939408872967)
        self.assertEqual(d._get_lambda(54, 1), 3.1514300233157844)
        self.assertEqual(d._get_lambda(54, 2), 3.1438896850317302)
        self.assertEqual(d._get_lambda(54, 3), 3.1361649560574856)
        self.assertEqual(d._get_lambda(54, 4), 3.1282473343306387)
        self.assertEqual(d._get_lambda(54, 5), 3.1201277383143937)
        self.assertEqual(d._get_lambda(54, 6), 3.1117964542894945)
        self.assertEqual(d._get_lambda(54, 7), 3.1032430776016966)
        self.assertEqual(d._get_lambda(54, 8), 3.0944564470227012)
        self.assertEqual(d._get_lambda(54, 9), 3.0854245712422848)

    def test_detect(self):
        data = [-0.25, 0.68, 0.94, 1.15, 1.20, 1.26, 1.26,
                1.34, 1.38, 1.43, 1.49, 1.49, 1.55, 1.56,
                1.58, 1.65, 1.69, 1.70, 1.76, 1.77, 1.81,
                1.91, 1.94, 1.96, 1.99, 2.06, 2.09, 2.10,
                2.14, 2.15, 2.23, 2.24, 2.26, 2.35, 2.37,
                2.40, 2.47, 2.54, 2.62, 2.64, 2.90, 2.92,
                2.92, 2.93, 3.21, 3.26, 3.30, 3.59, 3.68,
                4.30, 4.64, 5.34, 5.42, 6.01]
        exp = np.array([False] * 51 + [True] * 3)

        d = pydetect.GESDDetector(alpha=0.05)
        res = d.detect(data)
        tm.assert_numpy_array_equal(res, exp)

    def test_decompose(self):
        ap = pydetect.datasets.get_airpassengers()
        ap[50] += 100

        d = pydetect.GESDDetector(alpha=0.05, decompose=True)
        res = d.detect(ap)

        exp = np.zeros(144)
        exp[50] = 1.0
        exp = pd.Series(exp, index=ap.index)
        tm.assert_series_equal(res, exp)



#!/usr/bin/env python
# coding: utf-8

from __future__ import division

import numpy as np
from scipy.stats import t, zscore

from pydetect.base import OutlierDetector


class GESDDetector(OutlierDetector):

    """
    http://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm
    """

    def __init__(self, alpha=0.05, max_outliers=None, decompose=False):
        self.alpha = alpha
        self.max_outliers = max_outliers
        super(GESDDetector, self).__init__(decompose=decompose)

    def detect(self, data):
        data = self._validate(data)

        indexer, _, _ = self.get_statistics(data)
        result = np.zeros(len(data))
        result[indexer] = 1
        result = self._wrap_result(data, result)
        return result

    def _get_lambda(self, n, i):
        # because of the loop condition, original formula's i
        # is replaced by i + 1
        df = n - i - 2

        p = 1 - self.alpha / (2 * (n - i))
        # Critical value
        critical = t.ppf(p, df=df)
        lambda_i = (((n - i - 1) * critical) /
                    (np.sqrt((df + critical ** 2) * (n - i))))
        return lambda_i

    def get_statistics(self, data):

        n = len(data)
        if self.max_outliers is None:
            max_outliers = int(np.floor(n // 2))
        else:
            max_outliers = self.max_outliers

        remainings = data.copy()
        outliers = np.zeros(len(data))

        # number of outliers
        n_outliers = 0

        r = []
        l = []

        for i in range(max_outliers):
            # s must be sample standard deviation
            z_score = np.abs(zscore(remainings, ddof=1))
            max_pos = np.argmax(z_score)
            r_i = z_score[max_pos]

            lambda_i = self._get_lambda(n, i)

            # ToDo: ?
            r.append(r_i)
            l.append(lambda_i)

            if np.isnan(r_i) or np.isnan(lambda_i):
                break
            elif r_i > lambda_i:
                n_outliers = i + 1

            # convert to bool indexer
            max_indexer = np.arange(len(remainings)) == max_pos

            flag = np.zeros(len(max_indexer))
            flag[max_indexer] = i + 1

            outliers[outliers == 0] = flag

            # update remainings
            remainings = remainings[~max_indexer]

        outliers = (outliers != 0) & (outliers <= n_outliers)

        return outliers, np.array(r), np.array(l)



#!/usr/bin/env python
# coding: utf-8

import numpy as np
import pandas as pd


class BaseDetector(object):

    def __init__(self, decompose=False):
        self.decompose = decompose

    def detect(self, data):
        raise NotImplementedError

    def get_statistics(self, data):
        raise NotImplementedError

    def _validate(self, data):
        """ validate data """
        self._original = data

        if self.decompose:
            data = self._decompose(data)

        data = np.asarray(data)
        if data.ndim > 2:
            raise ValueError('Input must be less than 2 dimentions')
        elif data.ndim == 2 and data.shape[1] > 1:
            raise ValueError('Input must be univariate')

        return data

    def _decompose(self, data):
        try:
            import statsmodels.api as sm
        except ImportError:
            msg = ('statsmodels >= 0.6.0 is required to perform '
                   'seasonal decomposition')
            raise ImportError(msg)

        # if not isinstance(data, pd.Series):
        #     raise ValueError('Input must be pd.Series')

        # ToDo:
        # - check DatetimeIndex and freq

        decomposed = sm.tsa.seasonal_decompose(data)
        resid = decomposed.resid

        self._decompose_indexer = pd.notnull(resid.values)
        resid = resid[self._decompose_indexer]
        return resid

    def _wrap_result(self, data, result):
        """ wrap result to be compat with data """

        if self.decompose:
            # pad NaN
            pad = np.zeros(len(self._original))
            pad[self._decompose_indexer] = result
            result = pad

        if isinstance(self._original, (pd.Series, pd.DataFrame)):
            index = self._original.index
            result = self._original._constructor(result, index=index)
        return result


class OutlierDetector(BaseDetector):
    pass


class ChangePointDetector(BaseDetector):
    pass


class SingleChangePointDetector(ChangePointDetector):

    def detect(self, data):
        data = self._validate(data)

        indexer, null, value = self.get_statistics(data)
        result = np.zeros(len(data))
        result[indexer] = 1
        result = self._wrap_result(data, result)
        return result



#!/usr/bin/env python

import pydetect

import pandas as pd
import pandas.util.testing as tm


class TestDatasets(tm.TestCase):

    def test_nile(self):
        nile = pydetect.datasets.get_nile()
        self.assertIsInstance(nile, pd.Series)
        self.assertEqual(nile.name, 'Nile')

        # do not overwrite
        new = pydetect.datasets.get_nile()
        self.assertIsNot(nile, new)

    def test_airpassengers(self):
        ap = pydetect.datasets.get_airpassengers()
        self.assertIsInstance(ap, pd.Series)
        self.assertEqual(ap.name, 'Air Passengers')

        # do not overwrite
        new = pydetect.datasets.get_airpassengers()
        self.assertIsNot(ap, new)



#!/usr/bin/env python

import pydetect

import numpy as np
import pandas as pd
import pandas.util.testing as tm


class TestMeanDetector(tm.TestCase):

    def test_nile_statistics(self):
        nile = pydetect.datasets.get_nile()

        d = pydetect.MeanDetector()
        i, n, v = d.get_statistics(nile)

        self.assertEqual(27, i)
        self.assertEqual(2835156.75, n)
        self.assertEqual(1597457.1944444478, v)

    def test_input_class(self):
        nile = pydetect.datasets.get_nile()
        d = pydetect.MeanDetector()

        exp = np.array([0.] * 100)
        exp[27] = 1.
        exp_index = pd.date_range('1871-01-01', freq='AS', periods=100)
        # name will be reset
        exp = pd.Series(exp, exp_index)

        res = d.detect(nile)
        tm.assert_series_equal(res, exp)

        res = d.detect(nile.to_frame())
        tm.assert_frame_equal(res, exp.to_frame())

        res = d.detect(np.array(nile))
        tm.assert_numpy_array_equal(res, exp.values)

        res = d.detect(nile.tolist())
        tm.assert_numpy_array_equal(res, exp.values)


class TestVarianceDetector(tm.TestCase):

    def test_nile(self):
        nile = pydetect.datasets.get_nile()

        d = pydetect.VarianceDetector()
        i, n, v = d.get_statistics(nile)

        self.assertEqual(46, i)
        self.assertEqual(1025.2437598632698, n)
        self.assertEqual(1013.6146385197355, v)

    def test_input_class(self):
        nile = pydetect.datasets.get_nile()
        d = pydetect.VarianceDetector()

        exp = np.array([0.] * 100)
        exp[46] = 1.
        exp_index = pd.date_range('1871-01-01', freq='AS', periods=100)
        # name will be reset
        exp = pd.Series(exp, exp_index)

        res = d.detect(nile)
        tm.assert_series_equal(res, exp)

        res = d.detect(nile.to_frame())
        tm.assert_frame_equal(res, exp.to_frame())

        res = d.detect(np.array(nile))
        tm.assert_numpy_array_equal(res, exp.values)

        res = d.detect(nile.tolist())
        tm.assert_numpy_array_equal(res, exp.values)


class TestMeanVarianceDetector(tm.TestCase):

    def test_nile(self):
        nile = pydetect.datasets.get_nile()

        d = pydetect.MeanVarianceDetector()
        i, n, v = d.get_statistics(nile)

        self.assertEqual(27, i)
        self.assertEqual(1025.2437598632698, n)
        self.assertEqual(967.68788456436801, v)

    def test_input_class(self):
        nile = pydetect.datasets.get_nile()
        d = pydetect.MeanVarianceDetector()

        exp = np.array([0.] * 100)
        exp[27] = 1.
        exp_index = pd.date_range('1871-01-01', freq='AS', periods=100)
        # name will be reset
        exp = pd.Series(exp, exp_index)

        res = d.detect(nile)
        tm.assert_series_equal(res, exp)

        res = d.detect(nile.to_frame())
        tm.assert_frame_equal(res, exp.to_frame())

        res = d.detect(np.array(nile))
        tm.assert_numpy_array_equal(res, exp.values)

        res = d.detect(nile.tolist())
        tm.assert_numpy_array_equal(res, exp.values)



#!/usr/bin/env python
# coding: utf-8

from __future__ import division

import numpy as np
from pydetect.base import SingleChangePointDetector


class MeanDetector(SingleChangePointDetector):

    def get_statistics(self, data):

        n = len(data)
        x = np.cumsum(data)
        x2 = np.cumsum(data ** 2)

        null = x2[-1] - (x[-1] ** 2 / n)

        denom = np.arange(1, n + 1)
        values = (x2 - x ** 2 / denom + (x2[-1] - x2) -
                  ((x[-1] - x) ** 2) / (n - denom))

        i = np.nanargmin(values)
        return i, null, values[i]


class VarianceDetector(SingleChangePointDetector):

    def get_statistics(self, data):

        mu = np.mean(data)

        n = len(data)
        x = np.cumsum((data - mu) ** 2)
        null = n * np.log(x[-1] / n)

        denom = np.arange(1, n + 1)
        denom_rev = n - denom

        sigma1 = x / denom
        sigma1[sigma1 <= 0] = 1e-10

        sigman = (x[-1] - x) / denom_rev
        sigman[sigman <= 0] = 1e-10

        values = denom * np.log(sigma1) + denom_rev * np.log(sigman)

        i = np.nanargmin(values)
        return i, null, values[i]


class MeanVarianceDetector(SingleChangePointDetector):

    def get_statistics(self, data):
        n = len(data)
        y = np.cumsum(data)
        y2 = np.cumsum(data ** 2)

        null = n * np.log((y2[-1] - (y[-1] ** 2 / n)) / n)

        denom = np.arange(1, n + 1)
        denom_rev = n - denom

        sigma1 = (y2 - (y ** 2 / denom)) / denom
        sigma1[sigma1 <= 0] = 1e-10

        sigman = ((y2[-1] - y2) - (y[-1] - y) ** 2 / denom_rev) / denom_rev
        sigman[sigman <= 0] = 1e-10

        values = denom * np.log(sigma1) + denom_rev * np.log(sigman)

        i = np.nanargmin(values)
        return i, null, values[i]



#!/usr/bin/env python
# coding: utf-8

from pydetect.changepoint import (MeanDetector, VarianceDetector,  # noqa
                                  MeanVarianceDetector)            # noqa
import pydetect.datasets as datasets                               # noqa
from pydetect.outlier import GESDDetector                          # noqa



# -*- coding: utf-8 -*-
#!/usr/bin/env python

import codecs
import os
from setuptools import setup, find_packages

PACKAGE = 'pydetect'
README = 'README.rst'
REQUIREMENTS = 'requirements.txt'

VERSION = '0.2.0'

def read(fname):
  # file must be read as utf-8 in py3 to avoid to be bytes
  return codecs.open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()

def write_version_py(filename=None):
    cnt = """\
version = '%s'
"""
    a = open(filename, 'w')
    try:
        a.write(cnt % VERSION)
    finally:
        a.close()

version_file = os.path.join(os.path.dirname(__file__), PACKAGE, 'version.py')
write_version_py(filename=version_file)

setup(name=PACKAGE,
      version=VERSION,
      description='Change point and anomaly detector for time-series',
      long_description=read(README),
      author='sinhrks',
      author_email='sinhrks@gmail.com',
      url='http://pydetect.readthedocs.org/en/stable',
      license = 'BSD',
      packages=find_packages(),
      install_requires=list(read(REQUIREMENTS).splitlines())
      )
